# Copyright 2021 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import symfit as sf
from symfit.core.minimizers import BFGS, NelderMead, SLSQP, Powell
import matplotlib.pyplot as plt
import matplotlib.pyplot as plt
import argparse

def fit_data(data,guess=10.0,use_err=False):

    x_data = data[:,0]
    y_data = 1- data[:,1]
    if use_err:
       data_err= data[:,2]

    x, y = sf.variables('x, y')
    pKa, n = sf.parameters('pKa, n')
    model = sf.Model({y: (10**(n*(pKa-x))+1)**-1.0})
    pKa.value=guess
    n.value=1
    if use_err:
       fit = sf.Fit(model, x=x_data, y=y_data,sigma_y=data_err ,minimizer=Powell, absolute_sigma=True)
    else:
       fit = sf.Fit(model, x=x_data, y=y_data,minimizer=Powell)
    result = fit.execute()

    print("pKa.....................................", result.value(pKa),'+/-', result.stdev(pKa))
    print("n.......................................", result.value(n),'+/-', result.stdev(n))
    print("Regression coefficent:................", result.r_squared,'\n')

    x_out = np.arange(min(x_data),max(x_data),10**-3.0)
    y_out = fit.model(x=x_out, **result.params)[0]

    return x_out ,y_out, result.value(pKa), result.stdev(pKa), result.r_squared,result.value(n), result.stdev(n)


parser = argparse.ArgumentParser(description='Fit and plot titration curves.')
parser.add_argument('-f', dest='name', type=str,help='filename')
parser.add_argument('-n',dest='n', type=int,help='number of titratable sides')

args = parser.parse_args()


data=np.loadtxt(args.name)
data[:,1] = data[:,1]/args.n
data[:,2] = data[:,2]/args.n

x, y, pKa, err, R, n, nerr = fit_data(data,guess=10,use_err=True)

plt.plot(x,y,label=str(round(pKa,2)),c="#0039A6")

plt.errorbar(data[:,0],1-data[:,1],yerr=data[:,2],ls=' ',
                 lw=2,marker='>',c="#0039A6",markersize='7')

plt.legend()
plt.xlabel('pH',fontsize='12')
plt.ylabel('degree of deprotonation',fontsize='12')
plt.tight_layout()
plt.show()



# Copyright 2021 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from pathlib import Path


def get_ph_values(ph_vals):
    """
    Parameters
    ----------
    ph_vals : list of strings
        Each element is either a single value or a range with increments (e.g., 3:8:0.5)

    Returns
    -------
    all_vals : list of strings
        Ordered unique pH values with 2 decimal precision

    """
    all_vals = []
    for ph_val in ph_vals:
        if ':' in ph_val:
            init, final, increment = [float(i) for i in ph_val.split(':')]
            vals = np.arange(init, final+1e-3, increment)
            all_vals.extend(vals)
        else:
            all_vals.append(float(ph_val))

    all_vals = sorted(list(set(all_vals)))
    all_vals = [f'{val:.2f}' for val in all_vals]

    return all_vals


def get_sim_names(mdp_files):
    sims = []
    for mdp_file in mdp_files:
        if mdp_file.suffix != '.mdp':
            raise NameError('Provided mdp file(s) does not have the ".mdp" extension.')
        sims.append(mdp_file.stem)
    return sims


def update_top_file(file_name, top_file, ph, prefix):
    """
    Prepend the pH value to the beginning of the topology file and adjust the directory
    of included ITP files.
    """

    new_top_dir = Path(f'{prefix}{ph}') / file_name
    new_top_file = f'#define pH{ph}\n\n'

    top_file = top_file.split('\n')
    for line in top_file:
        if '#include' in line.replace(" ", ""):
            path = Path(line.replace('"', "'").split("'")[1])
            new_path = '..' / path
            line = line.replace(str(path), str(new_path))
        new_top_file += f'{line}\n'

    new_top_dir.write_text(new_top_file)


def make_new_dirs(sims, ph, prefix):
    for sim in sims:
        Path(f'{prefix}{ph}/{sim}').mkdir(parents=True, exist_ok=True)


def create_bash_script(sims, mdp_files, top_file, gro_file, out_file, ph_vals, prefix):
    """
    Create a bash script for running all the simulations necessary for the titration.

    Parameters
    ----------
    sims : list of strings
    mdp_files : list of strings
    top_file : Path
    gro_file : Path
    out_file : Path
    ph_vals : list of strings
    prefix : string

    """
    with open(out_file, 'w') as file:
        file.write('#!/bin/bash\n\n')
        file.write('pHrange=(' + ' '.join(ph_vals) + ')\n\n')
        file.write('for pH in ${pHrange[*]}; do\n')
        for i, (sim, mdp_file) in enumerate(zip(sims, mdp_files)):

            if i == 0:
                cd = f'  cd {prefix}' + '${pH}/' + f'{sim}\n'
                gro_file = f'../../{gro_file}'
            else:
                cd = f'  cd ../{sim}\n'
                gro_file = f'../{sims[i-1]}/confout.gro'

            file.write(cd)
            file.write(f'  gmx grompp -f ../../{mdp_file} -c {gro_file} -p ../{top_file}\n')
            file.write('  gmx mdrun\n\n')

            if i == len(sims)-1:
                file.write('  cd ../..\ndone\n')


def titrate(args):
    ph_vals = get_ph_values(args.ph)
    sims = get_sim_names(args.mdp_file)
    top_file = args.top_file.read_text()

    for ph in ph_vals:
        make_new_dirs(sims, ph, args.prefix)
        update_top_file(args.top_file, top_file, ph, args.prefix)

    create_bash_script(sims, args.mdp_file, args.top_file, args.gro_file, args.out_file, ph_vals,
                       args.prefix)



# Copyright 2021 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import MDAnalysis as mda
from scipy.spatial.transform import Rotation
import random


BEAD_TYPES = {'water': ['DN', 'H+'],
              'acid': ['D', 'DP', 'H+'],
              'base': ['D', 'DP'],
              'acid-deprot': ['D', 'DP'],
              'base-prot': ['D', 'DP', 'H+']}


def get_random_vectors(n_randoms=1000, length=1.7):
    """
    Generate randomly rotated 3 points that all have 120 degree angle to the
    center point with a given vector length. This is done n_randoms times.
    0.8660254 ~= 0.75**0.5

    Parameters
    ----------
    length : float, optional
        Length of the vectors. The default is 1.7 Angstrom.

    Returns
    -------
    (n_randoms, 3, 3) Numpy array

    """

    vectors = [[1, 0, 0], [-0.5, 0.8660254, 0], [-0.5, -0.8660254, 0]]
    random_rots = Rotation.random(n_randoms)

    random_vecs = np.empty((n_randoms, 3, 3))
    for i, vec in enumerate(vectors):
        random_vecs[:, i, :] = random_rots.apply(vec)

    return random_vecs*length


class Particles():
    def __init__(self):
        self.resids = []
        self.resnames = []
        self.names = []
        self.indices = []
        self.coords = []

    def add_particle(self, resid, resname, name, coord):
        self.resids.append(resid)
        self.resnames.append(resname)
        self.names.append(name)
        self.coords.append(coord)

    def __len__(self):
        return len(self.resids)


def check_input(args):
    if len(args.bead_type) != len(args.sel):
        raise ValueError('Unmatching number of -bead and -sel was provided.')


def select_titratable_beads(uni, selection_commands):
    """
    Parameters
    ----------
    uni : MDAnalysis Universe object
    selection_commands : list of strings
        List of MDAnalysis selection commands

    Returns
    -------
    selections : 1D Numpy array
        Array with the selection index of each atom. -1 if not in any selection

    """

    selections = np.full(len(uni.atoms), -1)

    for i, selection_command in enumerate(selection_commands):
        idx = uni.select_atoms(selection_command).indices

        if len(idx) == 0:
            raise ValueError(f'Selection command "{selection_command}" contains zero particles. '
                             'Check your input.')
        if np.any(selections[idx] != -1):
            raise ValueError(f'Selection command "{selection_command}" contains particles from '
                             'a different selection. Check your input.')

        selections[idx] = i

    return selections


def add_extra_particles(beads, protons, atom, particle_list, random_vecs, prot_name):
    """
    Select a random vector and add each necessary particle to either the protons or
    beads object.

    Parameters
    ----------
    beads : Particles object

    protons : Particles object

    atom : MDAnalysis atom object

    particle_list : list
        all the necessary particles for the corresponding BEAD_TYPE
    random_vecs : (n_randoms, 3, 3) Numpy array

    """

    vectors = random_vecs[random.randint(0, 999)]

    for i, particle in enumerate(particle_list):
        if particle == 'H+':
            protons.add_particle(None, particle, prot_name, atom.position + vectors[i])
        else:
            beads.add_particle(atom.resid, atom.resname, particle, atom.position + vectors[i])

    return beads, protons


def create_titratable_universe(beads, protons, dimensions):
    """
    Combines all beads and protons to create the titratable MDAnalysis universe

    Parameters
    ----------
    beads : Particles object

    protons : Particles object

    dimensions : (3) Numpy array
        Dimensions of the unit cell (Angstrom)

    """

    n_atoms = len(beads)+len(protons)
    max_bead_resid = max(beads.resids)
    resids = np.append(beads.resids, 1 + np.arange(max_bead_resid, max_bead_resid+len(protons)))
    resnames = beads.resnames + protons.resnames
    names = beads.names + protons.names

    uni = mda.Universe.empty(n_atoms, trajectory=True, n_residues=n_atoms,
                             atom_resindex=np.arange(0, n_atoms),
                             residue_segindex=np.zeros(n_atoms))

    uni.add_TopologyAttr('resid', resids)
    uni.add_TopologyAttr('resname', resnames)
    uni.add_TopologyAttr('name', names)

    uni.atoms.positions = np.append(beads.coords, protons.coords, axis=0)
    uni.dimensions = dimensions

    return uni


def conv_coords(args):
    check_input(args)

    beads = Particles()
    protons = Particles()
    random_vecs = get_random_vectors()

    uni = mda.Universe(args.input_file)
    selections = select_titratable_beads(uni, args.sel)

    for atom, selection in zip(uni.atoms, selections):
        beads.add_particle(atom.resid, atom.resname, atom.name, atom.position)
        if selection >= 0:
            particle_list = BEAD_TYPES[args.bead_type[selection]]
            beads, protons = add_extra_particles(beads, protons, atom, particle_list, random_vecs,
                                                 args.prot)

    titr_uni = create_titratable_universe(beads, protons, uni.dimensions)
    titr_uni.atoms.write(args.out_file)



# Copyright 2021 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import MDAnalysis as mda
from tqdm import tqdm
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from pathlib import Path
from .titrate import get_ph_values


def setup_dirs(ph_values, sub_dir, prefix):
    """
    Sets up directory paths based on whether multidir options have been provided or not.
    """

    if ph_values:
        ph_values = get_ph_values(ph_values)
        sim_dirs = [Path(f'{prefix}{ph}/{sub_dir}') for ph in ph_values]
    else:
        sim_dirs = [Path.cwd()]

    return sim_dirs, np.array(ph_values, dtype=float)


def save_degree_of_prot_to_file(out_file, ph_values, props):

    fmt = '%10.5f'*props.shape[1]
    header = ''.join([f'sel. {i}'.rjust(10) for i in range(1, props.shape[1]+1)])

    if ph_values.size != 0:
        props = np.column_stack((ph_values, props))
        fmt = '%6.2f' + fmt
        header = '    pH' + header
    fmt = '  ' + fmt

    np.savetxt(out_file, props, fmt=fmt, header=header)


def do_statistics(degrees_of_deprot):
    """
    This is probably where a more complicated statistical analysis would be inferfaced
    """

    means = np.array(degrees_of_deprot).mean(axis=-1)
    stds = np.array(degrees_of_deprot).std(axis=-1)

    return means, stds


def calc_degree_of_deprot(ph, pka, q):
    return 1/(10**(q*(pka-ph))+1)


def fit_titration_curve(ph_values, means, stds):
    fits = []
    fit_std_errs = []
    for mean, std in zip(means.T, stds.T):
        popt, pcov = curve_fit(calc_degree_of_deprot, ph_values, mean, p0=(7.5, 0.75),
                               sigma=std, absolute_sigma=True)
        std_err = np.sqrt(np.diag(pcov))
        fits.append(popt)
        fit_std_errs.append(std_err)
    return fits, fit_std_errs


def write_pka_to_file(output_file, ph_values, fits, fit_std_errs):
    sel_range = np.arange(1, len(fits)+1)
    fmt = '%8d' + '%10.5f'*4
    header = 'Sel. #       pKa         q  err(pKa)    err(q)'
    props = np.column_stack((sel_range, fits, fit_std_errs))
    np.savetxt(output_file, props, fmt=fmt, header=header)


def do_plotting(out_file, ph_values, fits, fit_std_errs, means):
    ph_range = np.arange(ph_values.min(), ph_values.max()+1e-3, 0.1)
    fig = plt.figure()

    for i, (fit, std_err, mean) in enumerate(zip(fits, fit_std_errs, means.T), start=1):
        points, = plt.plot(ph_values, mean, '^',
                           label=f'Sel. {i}, pKa = {fit[0]:.2f}\u00B1{std_err[0]:.2f}'
                                 f', q = {fit[1]:.2f}\u00B1{std_err[1]:.2f}')
        plt.plot(ph_range, calc_degree_of_deprot(ph_range, *fit), color=points.get_color())

    plt.xlabel('pH')
    plt.ylabel('Degree of deprotonation')
    plt.legend()
    fig.tight_layout()
    fig.savefig(out_file, bbox_inches='tight')


def construct_reference(u, sel_commands, ref_command):
    """
    Reference is constructed as ref_command + all selections that are not the current one.

    """

    if ref_command:
        refs = [u.select_atoms(ref_command) for _ in enumerate(sel_commands)]
    else:
        refs = [u.select_atoms('') for _ in enumerate(sel_commands)]

    if len(sel_commands) > 1:
        other_sels = [' or '.join([command for command in sel_commands if command != sel_command])
                      for sel_command in sel_commands]
        refs = [ref + u.select_atoms(other_sel) for ref, other_sel in zip(refs, other_sels)]

    return refs


def do_one_trajectory(sim_dir, tpr_file, traj_file, sel_commands, ref_command, prot, start, end,
                      scheme):
    """

    Parameters
    ----------
    sim_dir : Pathlib path
    tpr_file : Pathlib path
    traj_file : Pathlib path
    sel_commands : list of str
        All selection commands given by -sel
    ref_command : str
        Selection command for the remaining titratable beads, given by -ref
    prot : str
        name of the proton bead
    start : int
        time (ps) to start the analysis
    end : int
        time (ps) to end the analysis
    scheme : str
        proton counting scheme

    Returns
    -------
    degree_of_deprot : (n_selections, n_frames) np.array
        Degree of deprotonation matrix for each selection and for each frame

    """

    u = mda.Universe(str(sim_dir / tpr_file), str(sim_dir / traj_file))

    sels = [u.select_atoms(sel_command) for sel_command in sel_commands]
    prot = u.select_atoms(f'name {prot}')
    refs = construct_reference(u, sel_commands, ref_command)

    count = 0
    n_frames = len(u.trajectory[start:end])
    n_prot_matrix = np.zeros((len(sels), n_frames))
    time = np.zeros(n_frames)

    for ts in tqdm(u.trajectory[start:end]):
        time[count] = ts.time
        for i, (sel, ref) in enumerate(zip(sels, refs)):
            n_prot_matrix[i, count] = do_one_analysis(u, sel, ref, prot, scheme)
        count = count+1

    sel_sizes = np.array([len(sel) for sel in sels])
    degree_of_deprot = 1 - (n_prot_matrix.T/sel_sizes).T

    return degree_of_deprot


def do_one_analysis(u, sel, ref, prot, scheme):
    n_prots = 0
    for idx in sel.indices:

        curr = u.atoms[idx]
        oidx = sel.indices[sel.indices != idx]
        others = ref + u.atoms[oidx]

        # extract all water molecules close to the acid
        pairs, dists = mda.lib.distances.capped_distance(
            others.positions, curr.position, max_cutoff=11.0, box=u.dimensions)
        indices = np.where(dists > 0)
        water_less = others[pairs[indices][:, 0]]

        if len(water_less) != 0:
            # extract all protons close to acid
            pairs, dists = mda.lib.distances.capped_distance(
                prot.positions, curr.position, max_cutoff=11.0, box=u.dimensions)
            indices = np.where(dists > 0)
            prot_less = prot[pairs[indices][:, 0]]

            # compute all distances between water, acid and protons
            pairs_water, dists_water = mda.lib.distances.capped_distance(
                prot_less.positions, water_less.positions, max_cutoff=11, box=u.dimensions)
            pairs_acid, dists_acid = mda.lib.distances.capped_distance(
                prot_less.positions, curr.position, max_cutoff=11, box=u.dimensions)

            dist_mat_water = np.full((len(prot_less) + 1, len(water_less)), np.inf)
            dist_mat_water[pairs_water[:, 0] + 1, pairs_water[:, 1]] = dists_water

            dist_mat_acid = np.full((len(prot_less) + 1, 1), np.inf)
            dist_mat_acid[pairs_acid[:, 0] + 1, pairs_acid[:, 1]] = dists_acid

            min_water = dist_mat_water.min(axis=1)
            n_prot = dist_mat_acid[:, 0] < min_water

            if scheme == "legacy":
                if n_prot.sum() > 0:
                    n_prots += 1
            elif scheme == 'total':
                if n_prot.sum() > 0:
                    n_prots += n_prot.sum()

        else:
            pairs_acid, dists_acid = mda.lib.distances.capped_distance(
                prot_less.positions, curr.position, max_cutoff=11, box=u.dimensions)
            n_prot = dists_acid
            if len(n_prot) != 0:
                n_prots += 1

    return n_prots


def analyze(args):

    degrees_of_deprot = []
    sim_dirs, ph_values = setup_dirs(args.ph, args.dir, args.prefix)

    for sim_dir in sim_dirs:
        degree_of_deprot = do_one_trajectory(sim_dir, args.tpr, args.traj, args.sel, args.ref,
                                             args.prot, args.start, args.end, args.scheme)
        degrees_of_deprot.append(degree_of_deprot)

    means, stds = do_statistics(degrees_of_deprot)
    save_degree_of_prot_to_file(args.out, ph_values, means)

    if args.std_out:
        save_degree_of_prot_to_file(args.std_out, ph_values, stds)

    if args.fit:
        if ph_values.size < 4:
            print('WARNING: You requested a titration curve fitting with less than 4 pH data '
                  ' points. It will be ignored.')
        else:
            fits, fit_std_errs = fit_titration_curve(ph_values, means, stds)
            write_pka_to_file(args.fit_out, ph_values, fits, fit_std_errs)
            do_plotting(args.plot_out, ph_values, fits, fit_std_errs, means)
